/*
 * targa.datatype
 * (c) Fredrik Wikstrom
 */

/**********************************************************************/

//#define DEBUGMETHODS 0

/**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <exec/types.h>
#include <exec/memory.h>
#include <dos/dostags.h>
#include <graphics/gfxbase.h>
#include <graphics/gfx.h>
#include <graphics/rpattr.h>
#include <intuition/imageclass.h>
#include <intuition/icclass.h>
#include <intuition/gadgetclass.h>
#include <intuition/cghooks.h>
#include <datatypes/datatypesclass.h>
#include <datatypes/pictureclass.h>

#include <clib/alib_protos.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#include <proto/utility.h>
#include <proto/iffparse.h>
#include <proto/datatypes.h>

#include <aros/symbolsets.h>

ADD2LIBS("datatypes/picture.datatype", 0, struct Library *, PictureBase);

#include "debug.h"
#include "methods.h"
#include "targaclass.h"

#define RGB8to32(RGB) (((ULONG)(RGB) << 24)|((ULONG)(RGB) << 16)|((ULONG)(RGB) << 8)|((ULONG)(RGB)))

static LONG WriteTarga (Class *cl, Object *o, struct dtWrite *msg);
static LONG ConvertTarga (Class *cl, Object *o, BPTR file, struct BitMapHeader *bmh);
//static LONG GetTarga (Class *cl, Object *o, struct TagItem *tags);
static BOOL GetTarga (Class *cl, Object *o);

static LONG ReadLUT8 (Class *cl, Object *o, BPTR file, struct BitMapHeader *bmh, struct TargaHeader *tga);
static LONG ReadRGB16 (Class *cl, Object *o, BPTR file, struct BitMapHeader *bmh, struct TargaHeader *tga);
static LONG ReadTruecolor (Class *cl, Object *o, BPTR file, struct BitMapHeader *bmh, struct TargaHeader *tga);
static LONG ReadPixels (struct Library *libBase, struct TargaHeader *tga, BPTR file, UBYTE *buf, LONG width, LONG bpp);
//static LONG ReadRLE (struct Library *libBase, struct TargaHeader *tga, BPTR file, UBYTE *buf, LONG width, LONG bpp);
static LONG WriteLUT8 (Class *cl, Object *o, BPTR file, struct BitMapHeader *bmh, BOOL tgacomp);
static LONG WriteRGB16 (Class *cl, Object *o, BPTR file, UBYTE *linebuf, struct BitMapHeader *bmh, BOOL tgacomp);
static LONG WriteTruecolor (Class *cl, Object *o, BPTR file, UBYTE *linebuf, struct BitMapHeader *bmh, BOOL tgacomp);
static LONG WriteRLE (BPTR file, UBYTE *inBuf, int len, LONG bpp);
static LONG ConvertCmap (UBYTE *inbuf, UBYTE *outbuf, LONG len, LONG bytes);
static LONG GetCMAP24 (struct Library *libBase, BPTR file, struct ColorRegister *cmap, LONG len, LONG bytes);
static LONG SetCMAP24 (BPTR file, struct ColorRegister *colormap, UBYTE *CMAP, LONG len);
static BOOL CompareBytes(UBYTE *pix1, UBYTE *pix2, int len);
static LONG WriteBytes (BPTR file, char *data, LONG offset, LONG length);
    
/****************************************************************************************/

IPTR TARGA__OM_NEW(Class *cl, Object *o, Msg msg)
{
    Object *newobj;
    
    D(bug("[targa.datatype] %s()\n", __func__));
    
    newobj = (Object *)DoSuperMethodA(cl, o, msg);
    if (newobj)
    {
        if (!GetTarga(cl, newobj))
        {
            CoerceMethod(cl, newobj, OM_DISPOSE);
            newobj = NULL;
        }
    }

    return (IPTR)newobj;
}
/**************************************************************************************************/

//IPTR TARGA__OM_NEW(Class *cl, Object *o, struct opSet *msg)
//{
	//IPTR ret;
	//ret = DoSuperMethodA(cl, o, msg);
	//if (ret) {
		//LONG error;
		//error = GetTarga(cl, (Object *)ret, msg->ops_AttrList);
		//if (error != OK) {
			//CoerceMethod(cl, (Object *)ret, OM_DISPOSE);
			//ret = 0;
			//SetIoErr(error);
		//}
	//}
	//return ret;
//}

/**************************************************************************************************/

IPTR TARGA__DTM_WRITE(Class *cl, Object *o, struct dtWrite *dtw)
{
	IPTR ret;
	/* check if dt's native format should be used */
	if (dtw->dtw_Mode == DTWM_RAW) 
	{
		LONG error;
		error = WriteTarga(cl, o, dtw);
		if (error == OK) {
			ret = TRUE;
		} else {
			SetIoErr(error);
			ret = FALSE;
		}
	} else {
		ret = DoSuperMethodA(cl, o, dtw);
	}
	return ret;
    //return TRUE;
}

/**************************************************************************************************/

BOOL CompareBytes(UBYTE *pix1, UBYTE *pix2, int len)
{
	int i;
    if (pix1 == NULL || pix2 == NULL)
        return FALSE;
    
    for (i = 0; i < len; i++)
    {
        if (pix1[i] != pix2[i])
            return FALSE;
    }

    return TRUE;
}

/**************************************************************************************************/

LONG WriteBytes(BPTR file, char *data, LONG offset, LONG length)
{    
    LONG count = 0;

    Seek(file,offset,OFFSET_BEGINNING);    	
    count = Write(file, data, length);
    if (count != length)
    {	
	    return 0;
    }

    return count;
}

/**************************************************************************************************/

static LONG WriteTarga (Class *cl, Object *o, struct dtWrite *dtw)
{
    BPTR filehandle;
    int width, height, depth;	
	BOOL tgacomp = FALSE;
	LONG status, error = OK;    
	ULONG *cregs;
    UBYTE *linebuf;
    UBYTE tempRGB;
	ULONG numcolors = 0;    
    struct BitMapHeader *bmhd;
	struct ColorRegister *cmap;
        

	D(bug("targa.datatype/SaveTarga()\n"));

	/* A NULL file handle is a NOP */
	if( !dtw->dtw_FileHandle )
	{
		D(bug("targa.datatype/SaveTarga() --- empty Filehandle - just testing\n"));
		//return TRUE;
		error = OK;
		return error;
	}
	filehandle = dtw->dtw_FileHandle;
    
    
    /* GET DATATYPE ATTRIBUTES FROM DTO */    

	/* Get BitMapHeader */
	if( GetDTAttrs( o,  PDTA_BitMapHeader, (IPTR) &bmhd,                                
				TAG_DONE ) != 1UL ||
			!bmhd )
	{
		D(bug("targa.datatype/SaveTarga() --- missing attributes\n"));
		//SetIoErr(ERROR_OBJECT_WRONG_TYPE);
		//return FALSE;
		error = ERROR_NOT_IMPLEMENTED;
		return error;
	}
    
    /* Get Width, Height & BitDepth of Image.*/	
    width = bmhd->bmh_Width;
    height = bmhd->bmh_Height;
    depth = bmhd->bmh_Depth;
    
	/* disallow images with depth not eq to 8, 16, 24 or 32 */
    if (depth != 8 && depth != 16 && depth != 24 && depth != 32)    
	{
		D(bug("targa.datatype/SaveTarga() --- color depth %d, can save only depths of 8, 16, 24 or 32", depth));
		//SetIoErr(ERROR_OBJECT_WRONG_TYPE);
		//return FALSE;
		error = ERROR_NOT_IMPLEMENTED;
		return error;
	}
	D(bug("targa.datatype/SaveTarga() --- Picture size %d x %d (x %d bit)\n", width, height, depth));



    /* Set 16bit for testing */
    //if(bmhd->bmh_Depth == 24) 
	//{
		//bmhd->bmh_Depth = 16;
		//depth = 16;
	//}

	D(bug("targa.datatype/SaveTarga() --- Setup Pixel Data by BitDepth.\n"));

	/* Set Default Data for 24bit */
	int cmaptype = 0;
	int imagetype = 2;
	int cmapdepth = 0;
	int compressed = 0;
	//int compressed = 1;	
    ULONG pixelfmt = PBPAFMT_RGB;	
    ULONG alignwidth = (width * 3);

	if( depth == 8 )
	{
		cmaptype = 1;
		imagetype = 1;
		cmapdepth = 24;
		numcolors = 256;
		pixelfmt = PBPAFMT_LUT8;
		alignwidth = (width);		
	}
	else if( depth == 16 )
	{
		/* For 16bit, 24bit & 32bit there is no colormap */		
        /* For 16bit, Convert 24bit pixel Data to 16bit Data */
		/* Same pixel format as 24bit. pixelfmt = PBPAFMT_RGB */

		D(bug("targa.datatype/SaveTarga() --- 16bit mode - just testing\n"));
		pixelfmt = PBPAFMT_RGB; //Same settings as for 24bit	
        alignwidth = (width * 2); //Set alignwidth for 16bit		

		//See: http://www.fileformat.info/format/tga/
	    //Reading and writing 16-bit pixel data.            
	}
	else if( depth == 32 )
	{
        pixelfmt = PBPAFMT_RGBA;	
        alignwidth = (width * 4);	    
	}

    /* Set compression value */
	if(compressed == 1) 
	{
		imagetype += 8;
		tgacomp = TRUE;
	}
    

	/* PREPARE TARGA HEADER VALUES */         

     /* Set Values for new TargaHeader */
     UBYTE TGAHDR[18];
     TGAHDR[0] = 0;
     TGAHDR[1] = cmaptype;
     TGAHDR[2] = imagetype;
     TGAHDR[3] = TGAHDR[4] = 0;    
	 TGAHDR[5] = (UBYTE)(numcolors & 0xFF);
     TGAHDR[6] = (UBYTE)(numcolors >> 8);  
     TGAHDR[7] = cmapdepth; 
     TGAHDR[8] = TGAHDR[9] = 0;    
     TGAHDR[10] = TGAHDR[11] = 0; 
     TGAHDR[12] = (UBYTE)(width & 0xFF);
     TGAHDR[13] = (UBYTE)(width >> 8);    
     TGAHDR[14] = (UBYTE)(height & 0xFF);
     TGAHDR[15] = (UBYTE)(height >> 8); 
     TGAHDR[16] = depth; 
     TGAHDR[17] = 0;
	
     
     /* WRITE TARGA HEADER INFORMATION TO FILE */	 
    
	/* Write new targa header to File */
    if( FWrite( filehandle, TGAHDR, 18, 1 ) != 1 )
    {
        D(bug("targa.datatype/SaveTarga() --- writing targa header failed !\n"));            
        //return FALSE;
		error = NOTOK;
		goto out;
    }
	D(bug("targa.datatype/SaveTarga() --- writing targa header successful.\n"));


	/* WRITE TARGA COLORMAP DATA TO FILE */

    if( depth == 8 )
	{
		/* Convert ColorMap to CMAP Byte Array */		
		UBYTE Cmap[numcolors*3];
		if (GetDTAttrs(o, PDTA_ColorRegisters, &cmap, PDTA_CRegs, &cregs, TAG_END) == 2)
		{            			
			//SetColorMap24(filehandle, cmap, Cmap, numcolors);
			SetCMAP24(filehandle, cmap, Cmap, numcolors);
        }		
        		    
		/* Write CMAP to File */
		if( FWrite( filehandle, Cmap, numcolors*3, 1 ) != 1 )
        {
            D(bug("targa.datatype/SaveTarga() --- writing colormap data failed!\n"));            
            //return FALSE;
			error = NOTOK;
		    goto out;
		}
		D(bug("targa.datatype/SaveTarga() --- writing colormap data successful.\n"));
	}

     
	/* ALLOCATE MEMORY BUFFER */
    
    D(bug("targa.datatype/SaveTarga() --- ScanLine AllocVec to Store Pixel Data\n"));    
    //if( !(linebuf = AllocVec(alignwidth, MEMF_ANY)) )	
    //{
        //SetIoErr(ERROR_NO_FREE_STORE);
        //return FALSE;
		//error = ERROR_NO_FREE_STORE;
		//return error;
    //}

    /* Allocate memory for scanline data */ 
	linebuf = AllocMem(alignwidth, MEMF_ANY);
	if(!linebuf)
	{        
		error = ERROR_NO_FREE_STORE;
		return error;
    }


    /* WRITE SCANLINE PIXEL DATA TO FILE */

    D(bug("targa.datatype/SaveTarga() --- begin copying picture data with READPIXELARRAY\n"));

	    
    /* Process scanline data as needed before writing */			
	if(depth == 8)
	{	
		D(bug("targa.datatype/SaveTarga() --- writing to file (8bit mode).\n"));            			            			
		status = WriteLUT8(cl, o, filehandle, bmhd, tgacomp);
		if(status != OK)		
		{
			//SetIoErr(ERROR_OBJECT_WRONG_TYPE);
		    //return FALSE;
			error = IoErr();
			goto out;
		}						
	}
	else if(depth == 16)
	{		
		D(bug("targa.datatype/SaveTarga() --- writing to file (16bit mode).\n"));            			            			
		status = WriteRGB16(cl, o, filehandle, linebuf, bmhd, tgacomp);
		if(status != OK)		
		{
			//SetIoErr(ERROR_OBJECT_WRONG_TYPE);
		    //return FALSE;
			error = IoErr();
			goto out;
		}						
	}		
	else if((depth == 24) || (depth == 32))
    { 
		D(bug("targa.datatype/SaveTarga() --- writing to file (24/32bit mode).\n"));            			            			
		status = WriteTruecolor(cl, o, filehandle, linebuf, bmhd, tgacomp);
		if(status != OK)		
		{
			//SetIoErr(ERROR_OBJECT_WRONG_TYPE);
		    //return FALSE;
			error = IoErr();
			goto out;
		}
    }


    D(bug("targa.datatype/SaveTarga() --- Normal Exit\n"));
	//FreeVec(linebuf);
    //SetIoErr(0);

	out:
	/* Free Resources */
	FreeMem(linebuf, alignwidth);

    return error;       
}

/**************************************************************************************************/

static LONG ConvertTarga (Class *cl, Object *o, BPTR file, struct BitMapHeader *bmh) {
	struct Library *libBase = (struct Library *)cl->cl_UserData;
	LONG status, error = OK;
	struct TargaHeader tga;
	struct TargaHeaderFile tgafile;
	LONG (*ReadFunc) (Class *cl, Object *o, BPTR file, struct BitMapHeader *bmh, struct TargaHeader *tga);
	int needs_palette;

	status = Read(file, &tgafile, sizeof(struct TargaHeaderFile));
	if (status != sizeof(struct TargaHeaderFile)) {
		return ReadError(status);
	}

	tga.idsize = tgafile.idsize;
	tga.cmaptype = tgafile.cmaptype;
	tga.imagetype = tgafile.imagetype;
	tga.cmapstart = AROS_LE2WORD(*(WORD *)tgafile.cmapstart);

	tga.cmaplength = AROS_LE2WORD(*(WORD *)tgafile.cmaplength);
	tga.cmapdepth = tgafile.cmapdepth;
	tga.xstart = AROS_LE2WORD(tgafile.xstart);
	tga.ystart = AROS_LE2WORD(tgafile.ystart);
	tga.width = AROS_LE2WORD(tgafile.width);
	tga.height = AROS_LE2WORD(tgafile.height);
	tga.depth = tgafile.depth;
	tga.descriptor = tgafile.descriptor;

	bmh->bmh_Width = tga.width;
	bmh->bmh_Height = tga.height;
	bmh->bmh_Depth = tga.depth;

	if (tga.cmaptype > 1) {
		return ERROR_NOT_IMPLEMENTED;
	}

	needs_palette = FALSE;
	ReadFunc = NULL;
	switch (tga.imagetype & ~TGA_IMAGE_RLE) {

		case TGA_IMAGE_LUT:
			if (tga.depth == 8) {
				needs_palette = TRUE;
				ReadFunc = ReadLUT8;
			}
			break;

		case TGA_IMAGE_GREY:
			if (tga.depth == 8) {
				ReadFunc = ReadLUT8;
			}
			break;

		case TGA_IMAGE_RGB:
			switch (tga.depth) {
				
				case 16:
					//bmh->bmh_Depth = 24;
					ReadFunc = ReadRGB16;
					break;

				case 24:
				case 32:
					ReadFunc = ReadTruecolor;
					break;

			}
			break;

	}

	if (!ReadFunc) {
		error = DTERROR_UNKNOWN_COMPRESSION;
		goto out;
	}

	if (tga.idsize) {
		if (Seek(file, tga.idsize, OFFSET_CURRENT) == -1 && (error = IoErr())) {
			goto out;
		}
	}

	if (tga.cmaptype > 1) {
		error = NOTOK;
		goto out;
	}
	if (needs_palette) {
		struct ColorRegister *cmap = NULL;
		ULONG *cregs = NULL;		
		LONG (*CMAPFunc) (struct Library *libBase, BPTR file, struct ColorRegister *cmap,
		    LONG len, int bytes);

		if (tga.cmaptype == TGA_CMAP_NONE) {
			error = NOTOK;
			goto out;
		}

		CMAPFunc = NULL;
		switch (tga.cmapdepth) {
			case 24:
				CMAPFunc = GetCMAP24;
				break;
		}
		if (!CMAPFunc) {
			error = DTERROR_UNKNOWN_COMPRESSION;
			goto out;
		}

		SetDTAttrs(o, NULL, NULL,
			PDTA_NumColors, 256,
			TAG_END);

		if (GetDTAttrs(o, PDTA_ColorRegisters, &cmap, PDTA_CRegs, &cregs, TAG_END) == 2)
		{
			int i;			
			int start, len, bytes;

			start = tga.cmapstart;
			len = tga.cmaplength;
			bytes = tga.cmapdepth; //15,16,24,32
			
			error = CMAPFunc(libBase, file, cmap+=start, len, bytes);
			if (error) goto out;

			cregs+=start;
			for (i = 0; i < len; i++) {
				*cregs++ = RGB8to32(cmap->red);
				*cregs++ = RGB8to32(cmap->green);
				*cregs++ = RGB8to32(cmap->blue);
				cmap++;
			}
		}
	} else {
		ULONG len;
		len = tga.cmaplength * ((tga.cmapdepth+7)>>3);
		if (Seek(file, len, OFFSET_CURRENT) == -1 && (error = IoErr())) {
			goto out;
		}
	}

	error = ReadFunc(cl, o, file, bmh, &tga);

    out:
	return error;
}

LONG ReadLUT8 (Class *cl, Object *o, BPTR file, struct BitMapHeader *bmh, struct TargaHeader *tga) {
	struct Library *libBase = (struct Library *)cl->cl_UserData;
	ULONG bpr, fmt;
	UBYTE *buf;
	LONG status, level = 0, error = OK;
	int y;
	
	bpr = bmh->bmh_Width;
	fmt = ((tga->imagetype & ~TGA_IMAGE_RLE) == TGA_IMAGE_GREY) ? PBPAFMT_GREY8 : PBPAFMT_LUT8;
	
	buf = AllocMem(bpr, /* MEMF_SHARED */ MEMF_ANY);
	if (!buf) {
		return ERROR_NO_FREE_STORE;
	}
	
	SetDTAttrs(o, NULL, NULL,
		DTA_NominalHoriz,	bmh->bmh_Width,
		DTA_NominalVert,	bmh->bmh_Height,
		PDTA_SourceMode,	PMODE_V43,
		DTA_ErrorLevel,		&level,
		DTA_ErrorNumber,	&error,
		TAG_END);
	if (level)
		goto out;
	else
		error = OK;

	BOOL is_bottom_up = FALSE;
	
	/* Note: right-to-left order is not honored by some Targa readers */
	//tga.imagedesc: 00vhaaaa, h=flip_horiz, v=flip_vert, a=alpha_bits	
	int tga_flipHoriz = ((tga->descriptor & 0x10) >> 4 ); //tga.imagedesc bit4
	D(bug("\ntarga.datatype/LoadTarga(): tga_flipHoriz: %d\n", tga_flipHoriz));
	int tga_flipVert  = ((tga->descriptor & 0x20) >> 5 ); //tga.imagedesc bit5
	D(bug("\ntarga.datatype/LoadTarga(): tga_flipVert: %d\n", tga_flipVert));
	if(tga_flipVert == 0) is_bottom_up = TRUE; //hv=00 bottom_up, hv=01 top_down

    int yoffset;
	//for (y = bmh->bmh_Height-1; y >= 0; y--) 
	for (y = 0; y < bmh->bmh_Height; y++)
	{
        /* Flip scanline if bottom_up */  
		if(is_bottom_up == TRUE)
		    yoffset = (bmh->bmh_Height - y - 1);
		else
		    yoffset = y; //Then it's top-Down so don't flip.

		status = ReadPixels(libBase, tga, file, buf, bpr, 1);
		if (status != bmh->bmh_Width) {
			error = IoErr();
			goto out;
		}
		
		DoSuperMethod(cl, o,
			PDTM_WRITEPIXELARRAY, buf, fmt, bpr, 0, yoffset, bpr, 1);
	}

out:
	FreeMem(buf, bpr);
	return error;
}

LONG ReadRGB16 (Class *cl, Object *o, BPTR file, struct BitMapHeader *bmh, struct TargaHeader *tga) 
{
	/*
    ** Read 16-bit Scanline
    ** stored in a datatype object
    */	

	//See: http://www.fileformat.info/format/tga/
	//Reading and writing 16-bit pixel data.
	
	struct Library *libBase = (struct Library *)cl->cl_UserData;  
	
    int x, y;
	int offset;
	WORD pixel;       
    UBYTE *buf1, *buf2;
    ULONG bpr, rgb, fmt;
	BYTE red, green, blue;
	LONG retval, level = 0, error = OK;	

	fmt = PBPAFMT_RGB;
	bpr = bmh->bmh_Width * 2;
    rgb = bmh->bmh_Width * 3;
    
	buf1 = AllocMem(bpr, MEMF_ANY);
	buf2 = AllocMem(rgb, MEMF_ANY);    
	if (!buf1 || !buf2)    
	{	
        error = ERROR_NO_FREE_STORE;
		return error;		
	}
	
    D(bug("\ntarga.datatype/LoadTarga(): Beginning Reading. (16 bit mode)\n"));

	SetDTAttrs(o, NULL, NULL,
		DTA_NominalHoriz,	bmh->bmh_Width,
		DTA_NominalVert,	bmh->bmh_Height,
		PDTA_SourceMode,	PMODE_V43,
		DTA_ErrorLevel,		&level,
		DTA_ErrorNumber,	&error,
		TAG_END);
	if (level)
		goto exit;
	else
		error = OK;

	BOOL is_bottom_up = FALSE;

	/* Note: right-to-left order is not honored by some Targa readers */
	//tga.imagedesc: 00vhaaaa, h=flip_horiz, v=flip_vert, a=alpha_bits	
	int tga_flipHoriz = ((tga->descriptor & 0x10) >> 4 ); //tga.imagedesc bit4
	D(bug("\ntarga.datatype/LoadTarga(): tga_flipHoriz: %d\n", tga_flipHoriz));
	int tga_flipVert  = ((tga->descriptor & 0x20) >> 5 ); //tga.imagedesc bit5
	D(bug("\ntarga.datatype/LoadTarga(): tga_flipVert: %d\n", tga_flipVert));
	if(tga_flipVert == 0) is_bottom_up = TRUE; //hv=00 bottom_up, hv=01 top_down	        
     
	int yoffset;
	for (y = 0; y < bmh->bmh_Height; y++)	
	{      
		/* Flip scanline if bottom_up */  
		if(is_bottom_up == TRUE)
		    yoffset = (bmh->bmh_Height - y - 1);
		else
		    yoffset = y; //Then it's top-Down so don't flip.

		/* Read scanline into buffer (width*2) */
		retval = ReadPixels(libBase, tga, file, buf1, bmh->bmh_Width, 2);		
		if (retval != bmh->bmh_Width) 
		{
			error = IoErr();
			goto exit;
		}

        /* Condense two bytes (16bit) into WORD */
        for(x = 0; x < bmh->bmh_Width; x++)		
        {
            
            /* Swap BGR to RGB while bitshifting */            
			red = ((buf1[(x*2)+1] << 1) & 0xf8);
			red += red >> 5;
			green = ((buf1[x*2] & 0xe0) >> 2) | ((buf1[(x*2)+1] & 0x03) << 6);
			green += green >> 5;
            blue = ((buf1[x*2] << 3) & 0xf8);
			blue += blue >> 5;

            /* Assign new values to rgb line buffer */            
            buf2[x*3] = red; buf2[(x*3)+1] = green; buf2[(x*3)+2] = blue;
        }
                
		DoSuperMethod(cl, o,
			PDTM_WRITEPIXELARRAY, buf2, fmt, rgb, 0, yoffset, bmh->bmh_Width, 1);
    }	
        
	exit:
	/* Free Resources */
	if (buf1) FreeMem(buf1, bpr);
	if (buf2) FreeMem(buf2, rgb);

	return error;
}

LONG ReadTruecolor (Class *cl, Object *o, BPTR file, struct BitMapHeader *bmh, struct TargaHeader *tga) {
	struct Library *libBase = (struct Library *)cl->cl_UserData;
	ULONG bpp, bpr, fmt;
	UBYTE *buf1, *buf2, *src, *dst;
	LONG status;
	IPTR level = 0, error = OK;
	int y, x;
	
	switch (tga->depth) {
		case 24:
			bpp = 3;
			fmt = PBPAFMT_RGB;
			break;
		case 32:
			bpp = 4;
			fmt = PBPAFMT_ARGB;
			break;
		default:
			return ERROR_NOT_IMPLEMENTED;
			break;
	}
	bpr = bmh->bmh_Width * bpp;
		
	buf1 = AllocMem(bpr, /* MEMF_SHARED */ MEMF_ANY);
	buf2 = AllocMem(bpr, /* MEMF_SHARED */ MEMF_ANY);
	if (!buf1 || !buf2) {
		error = ERROR_NO_FREE_STORE;
		goto out;
	}
	
	SetDTAttrs(o, NULL, NULL,
		DTA_NominalHoriz,	bmh->bmh_Width,
		DTA_NominalVert,	bmh->bmh_Height,
		PDTA_SourceMode,	PMODE_V43,
		DTA_ErrorLevel,		&level,
		DTA_ErrorNumber,	&error,
		TAG_END);
	if (level)
		goto out;
	else
		error = OK;

    BOOL is_bottom_up = FALSE;

	/* Note: right-to-left order is not honored by some Targa readers */
	//tga.imagedesc: 00vhaaaa, h=flip_horiz, v=flip_vert, a=alpha_bits	
	int tga_flipHoriz = ((tga->descriptor & 0x10) >> 4 ); //tga.imagedesc bit4
	D(bug("\ntarga.datatype/LoadTarga(): tga_flipHoriz: %d\n", tga_flipHoriz));
	int tga_flipVert  = ((tga->descriptor & 0x20) >> 5 ); //tga.imagedesc bit5
	D(bug("\ntarga.datatype/LoadTarga(): tga_flipVert: %d\n", tga_flipVert));
	if(tga_flipVert == 0) is_bottom_up = TRUE; //hv=00 bottom_up, hv=01 top_down

    	
	int yoffset;
	for (y = 0; y < bmh->bmh_Height; y++)
	//for (y = bmh->bmh_Height-1; y >= 0; y--) 
	{
        /* Flip scanline if bottom_up */  
		if(is_bottom_up == TRUE)
		    yoffset = (bmh->bmh_Height - y - 1);
		else
		    yoffset = y; //Then it's top-Down so don't flip.

		status = ReadPixels(libBase, tga, file, buf1, bmh->bmh_Width, bpp);
		if (status != bmh->bmh_Width) {
			error = IoErr();
			goto out;
		}

		src = buf1;
		dst = buf2;
		if (bpp == 3) {
			for (x = 0; x < bmh->bmh_Width; x++) {
				*dst++ = src[2];
				*dst++ = src[1];
				*dst++ = src[0];
				src += 3;
			}
		} else {
			for (x = 0; x < bmh->bmh_Width; x++) {
				//write_le32(dst, *(ULONG *)src);
				// FIXME: do we need endianess conversion
				*dst++ = src[3];
				*dst++ = src[2];
				*dst++ = src[1];
				*dst++ = src[0];
				src += 4;
			}
		}

		DoSuperMethod(cl, o,
			PDTM_WRITEPIXELARRAY, buf2, fmt, bpr, 0, yoffset, bmh->bmh_Width, 1);		
	}	
	
out:
	if (buf1) FreeMem(buf1, bpr);
	if (buf2) FreeMem(buf2, bpr);
	return error;
}

LONG ReadPixels (struct Library *libBase, struct TargaHeader *tga, BPTR file, UBYTE *buf, LONG width, LONG bpp) {
	if (tga->imagetype & TGA_IMAGE_RLE) {
		//UBYTE *ptr = buf;
		LONG x = 0;
		LONG st, cnt, rem;
		UBYTE pix[4];
		while (x < width) {
			cnt = FGetC(file);
			if (cnt == -1) break;
			rem = width-x;
			if (cnt & 0x80) {
				cnt -= 0x80;
				if (++cnt > rem) cnt = rem;
				st = FRead(file, pix, bpp, 1);
				if (st != 1) break;
				x += cnt;
				switch (bpp) {
					case 1:
						while (cnt--) {
							*buf++ = pix[0];
						}
						break;
					case 2:
						while (cnt--) {
							*(UWORD *)buf = *(UWORD *)pix;
							buf += 2;
						}
						break;
					case 3:
						while (cnt--) {
							*buf++ = pix[0];
							*buf++ = pix[1];
							*buf++ = pix[2];
						}
						break;
					case 4:
						while (cnt--) {
							*(ULONG *)buf = *(ULONG *)pix;
							buf += 4;
						}
						break;
				}
			} else {
				if (++cnt > rem) cnt = rem;
				x += st = FRead(file, buf, bpp, cnt);
				if (st != cnt) break;
				buf += (cnt*bpp);
			}
		}
		return x;
	} else {
		return FRead(file, buf, bpp, width);
	}
}

LONG WriteLUT8(Class *cl, Object *o, BPTR file, struct BitMapHeader *bmh, BOOL tgacomp)
//LONG WriteLUT8(Class *cl, Object *o, BPTR file, struct BitMapHeader *bmh) 
{    
	/*
    ** Write 8-bit Scanline
    ** stored in a 8-bit image file
    */
    
	int y;
	int height;
	int depth;	
	int yoffset;			
	LONG error = OK;
	ULONG bpr, fmt;
	UBYTE *buf;

    fmt = PBPAFMT_LUT8;
	bpr = bmh->bmh_Width;
	height = bmh->bmh_Height;
	depth = bmh->bmh_Depth;
	yoffset = height - 1;

	buf = AllocMem(bpr, MEMF_ANY);
	if (!buf) 
	{
		error = ERROR_NO_FREE_STORE;
		return error;
	}

	/* Read ScanLine Data from Bottom to Top for TGA */    
	for (y=0; y<height; y++)
    {
		if(!DoSuperMethod(cl, o,
		    PDTM_READPIXELARRAY, (IPTR)buf, fmt, bpr, 0, yoffset, bmh->bmh_Width, 1))
		    {
                D(bug("targa.datatype/SaveTarga() --- READPIXELARRAY line %d failed !\n", y));                
				error = IoErr();
			    goto exit;
            }

        /* Write rle encoded scanline to the output file */
        if(tgacomp == TRUE) WriteRLE(file, buf, bpr, depth);

		//LONG lineSize = 0;
	    //LONG lineOffset = ??; //header size + colormap
		/* Write RLE Compressed Scanline Picture Data to File */				        
		//lineSize = WriteLine(filehandle, linebuf, alignwidth, depth);
		//lineOffset += lineSize;

        /* Write the raw scanline data to the output file */
        if( FWrite( file, buf, bpr, 1 ) != 1 )
        {
            D(bug("targa.datatype/SaveTarga() --- writing picture data line %d failed !\n", y));            
			error = IoErr();
			goto exit;
        }		
        yoffset--;
    }	

	exit:
	/* Free Resources */	
	if(buf) FreeMem(buf, bpr);

	return error;
}

LONG WriteRGB16 (Class *cl, Object *o, BPTR file, UBYTE *linebuf, struct BitMapHeader *bmh, BOOL tgacomp) 
{
    /*
    ** Write 16-bit Scanline
    ** stored in a 16-bit image file
    */	

	//See: http://www.fileformat.info/format/tga/
	//Reading and writing 16-bit pixel data.

	
    int x, y;
	int offset;
	int depth;
	BYTE r,g,b;	     
    WORD rgb555; 
    WORD rgb565;
	LONG error = OK;	      
    UBYTE *buf1, *buf2;
    ULONG bpr, rgb, fmt;
    BYTE red, green, blue;	

    offset = 0;	
    fmt = PBPAFMT_RGB;
	bpr = bmh->bmh_Width * 2;
    rgb = bmh->bmh_Width * 3;
	depth = bmh->bmh_Depth;    
	
	buf2 = AllocMem(rgb, MEMF_ANY);
	if (!buf2)     
	{	
        error = ERROR_NO_FREE_STORE;
		return error;		
	}

	
    D(bug("\ntarga.datatype/LoadTarga(): Beginning Writing. (16 bit mode)\n"));
    
	for (y = 0; y < bmh->bmh_Height; y++)		
	{
        /* Read scanline from datatype into buf2 (width*3) */
        if(!DoSuperMethod(cl, o,
			PDTM_READPIXELARRAY, (IPTR)buf2, fmt, rgb, 0, bmh->bmh_Height - y - 1, bmh->bmh_Width, 1))
        {
            D(bug("targa.datatype/SaveTarga() --- READPIXELARRAY line %d failed !\n", y));           
            error = IoErr();
			goto exit;
        }

		/* Write rle encoded scanline to the output file */
        if(tgacomp == TRUE) WriteRLE(file, buf2, rgb, depth);        
        
        /* Condense 3 bytes (24bit) into (16bit) WORD */
        for(x = 0; x < bmh->bmh_Width; x++)		
        {                       
			/* Collect 24bit pixel elements */
            red = (buf2[x*3]);
            green = (buf2[(x*3)+1]);
            blue = (buf2[(x*3)+2]);
            
			/* Swap BGR to RGB format while bitshifting */			
			linebuf[x*2] = ((green & 0xF8) << 2) | ((blue & 0xF8) >> 3); // Take remaining 3 Bits of Green component & 5 bits of Blue
            linebuf[(x*2)+1] = ((red & 0xF8) >> 1) | ((green & 0xC0) >> 6); // Take 5 bits of Red component & 2 bits of Green						
        }		
				
		/* Now write the processed scanline data to the output file */
        if( FWrite( file, linebuf, bpr, 1 ) != 1 )
        {
            D(bug("targa.datatype/SaveTarga() --- writing picture scanline %d failed !\n", y));                        
			error = IoErr();
			goto exit;
        }
    }	
	
    exit:
	/* Free Resources */	
	if (buf2) FreeMem(buf2, rgb);

	return error;
}

LONG WriteTruecolor(Class *cl, Object *o, BPTR file, UBYTE *linebuf, struct BitMapHeader *bmh, BOOL tgacomp)
{
    /*
    ** Write 24 or 32-bit Scanline
    ** stored in a 24/32bit image file
    */
    
	int y, idx;	
	int width;
	int height;
	int depth;
	int yoffset;
	UBYTE *buf;	
	UBYTE a,r,g,b;
	LONG error = OK;
	ULONG bpr, bpp, fmt;

    width = bmh->bmh_Width;
	height = bmh->bmh_Height;
	depth = bmh->bmh_Depth;
	yoffset = height - 1;


    if(bmh->bmh_Depth == 24)
	{
	    bpp = 3;
		fmt = PBPAFMT_RGB;		
	}
	else if(bmh->bmh_Depth == 32)
	{
	    bpp = 4;
		fmt = PBPAFMT_ARGB;		
	}
	else
	{
        error = ERROR_NOT_IMPLEMENTED;
		return error;
	}
    bpr = bmh->bmh_Width * bpp;

		
	buf = AllocMem(bpr, MEMF_ANY);
	if (!buf) 
	{
		error = ERROR_NO_FREE_STORE;
		return error;
	}


    D(bug("\ntarga.datatype/WriteTrueColor(): Beginning Writing. (24/32bit mode)\n"));

    /* Read ScanLine Data from Bottom to Top for TGA */    
	for (y=0; y<height; y++)
	//for (y=0; y<15; y++)
    {
		if(!DoSuperMethod(cl, o,
		    PDTM_READPIXELARRAY, (IPTR)buf, fmt, bpr, 0, yoffset, bmh->bmh_Width, 1))
		    {
                D(bug("targa.datatype/WriteTrueColor() --- READPIXELARRAY line %d failed !\n", y));                
				error = IoErr();
			    goto exit;
            }
		        

		/* Process scanline data */
		if (bpp == 3) 
		{						
			/* Reverse 'RGB' to 'BGR' for TGA file for 24bit */				
            for (idx = 0; idx < (bmh->bmh_Width * bpp); idx+=3)
			{
				r = buf[idx];
				g = buf[idx + 1];
				b = buf[idx + 2];
				linebuf[idx] = b;
				linebuf[idx + 1] = g;
				linebuf[idx + 2] = r;                
			}
		} 
		else if (bpp == 4)
		{			
			/* Reverse 'ARGB' to 'BGRA' for TGA file for 32bit */
            for (idx = 0; idx < (bmh->bmh_Width * bpp); idx+=4)
			{	 
				a = buf[idx];
				r = buf[idx + 1];
				g = buf[idx + 2];
				b = buf[idx + 3];
				linebuf[idx] = b;
				linebuf[idx + 1] = g;
				linebuf[idx + 2] = r;
				linebuf[idx + 3] = a;				
			}
		}
        
        if(tgacomp == TRUE)
		{
			if (bpp == 3)
		    {
			    /* Reverse 'RGB' to 'BGR' for TGA file for 24bit */				
                for (idx = 0; idx < (bmh->bmh_Width * bpp); idx+=3)
			    {
				    r = buf[idx];
				    g = buf[idx + 1];
				    b = buf[idx + 2];
				    buf[idx] = b;
				    buf[idx + 1] = g;
				    buf[idx + 2] = r;                
			    }
			}
			else if (bpp == 4)
		    {			
			    /* Reverse 'ARGB' to 'BGRA' for TGA file for 32bit */
                for (idx = 0; idx < (bmh->bmh_Width * bpp); idx+=4)
			    {	 
				    a = buf[idx];
				    r = buf[idx + 1];
				    g = buf[idx + 2];
				    b = buf[idx + 3];
				    linebuf[idx] = b;
				    linebuf[idx + 1] = g;
				    linebuf[idx + 2] = r;
				    linebuf[idx + 3] = a;				
			    }
		    }
			/* Write rle encoded scanline data to the output file */
			D(bug("targa.datatype/SaveTarga() --- writing rle data line: %d\n", y));
			WriteRLE(file, buf, bmh->bmh_Width, bpp); //1,2,3,4
		}
		else
		{
            /* Write unencoded scanline data to the output file */
            if( FWrite( file, linebuf, bpr, 1 ) != 1 )
            {
                D(bug("targa.datatype/SaveTarga() --- writing picture data line %d failed !\n", y));            
			    error = IoErr();
			    goto exit;
            }
        }
        yoffset--;
    }	

	exit:
	/* Free Resources */	
	if(buf) FreeMem(buf, bpr);

	return error;
}

LONG WriteRLE(BPTR file, UBYTE *inBuf, int len, LONG bpp)
{   
    int j, k, cnt;
    int offset = 0;	
	int runCount = 0;
    int dirCount = 0;	       
    UBYTE pixBuf1[4];
	UBYTE pixBuf2[4];	
	UBYTE outBuf[len*bpp*2];


	D(bug("targa.datatype/WriteRLE() --- Length of Data: %d\n", len));

	//D(bug("targa.datatype/SaveTarga() --- writing rle encoded scanline\n"));

	// If a scanline begins with a direct sequence or a run count, it gets skipped.
	// That happens when scanlines are not continuous. The data doesn't carry over.
	// The solution is to finish the last pixel sequence at the end of ea scanline.

	// Banding on photo-quality images with predominately lengthy direct sequences.
	// When a dirCount of 128 is written it has an incorrect offset 388+1*2 causing
	// a shift to the right but what is written are junk bytes causing the banding.
	 
    for (j = 0; j < len; j++)    
    {		  
		memcpy(pixBuf1, inBuf, bpp); //bpp = 1,2,3,4
		memcpy(pixBuf2, inBuf + bpp, bpp); 
		if(CompareBytes(pixBuf1, pixBuf2, bpp) == TRUE) //pixels match              
        {
            /* There is a run - next pixel is the same */
			
            if (dirCount > 0)
            {
				//D(bug("targa.datatype/WriteRLE() --- Begin Buffer Offset: %d\n", offset));
				//D(bug("targa.datatype/WriteRLE() --- Current DirCount: %d\n", dirCount-1));                
                outBuf[offset] = dirCount-1; //Raw pixel //write pixel				
				memcpy(outBuf+offset+1, inBuf-bpp*dirCount, bpp*dirCount);
                cnt = bpp*dirCount + 1;
                
				/* reset run */
                dirCount = 0;
                runCount = 1;
                offset += cnt;
				//D(bug("targa.datatype/WriteRLE() --- End Buffer Offset: %d\n", offset));
			}
			else
			{				
                runCount += 1;
			}                                            
        }
        else        
	    { 
            /* There is No run - next pixel is different */
			//D(bug("targa.datatype/WriteRLE() --- No Match. Pixel is different.\n"));
			//D(bug("targa.datatype/WriteRLE() --- New RunCount: %d\n", runCount));
						
            if (runCount > 0)
            {
				//D(bug("targa.datatype/WriteRLE() --- Begin Buffer Offset: %d\n", offset));
				//D(bug("targa.datatype/WriteRLE() --- Current RunCount: %d\n", runCount));
                outBuf[offset] = 0x80 | runCount;								
                memcpy(outBuf+offset+1, pixBuf1, bpp);								
                cnt = bpp + 1;

                /* reset run */                
                dirCount = 0;
				runCount = 0;
                offset += cnt;
            }
            else
            {
				//Raw pixels
                dirCount += 1;
				//D(bug("targa.datatype/WriteRLE() --- New DirCount: %d\n", dirCount));				
            }                       
        }
		inBuf += bpp;

		if (runCount == 128)
        {
			//D(bug("targa.datatype/WriteRLE() --- Current RunCount: %d\n", runCount));
            //RLE pixels //write pixels
            outBuf[offset] = 255;
            memcpy(outBuf+offset+1, pixBuf1, bpp);		               		                
		    cnt = bpp + 1;
                    				
			/* reset run */            
            dirCount = 0;
            runCount = 0;
            offset += cnt;
        }		
        else if (dirCount == 128)
        {
			//D(bug("targa.datatype/WriteRLE() --- Current DirCount: %d\n", dirCount-1));
            //Raw pixels //write pixels
            outBuf[offset] = 127;
            memcpy(outBuf+offset+1, inBuf-bpp*dirCount, bpp*dirCount);
            cnt = bpp*dirCount + 1;             

            /* reset run */            
            dirCount = 0;
            runCount = 0;
            offset += cnt;
		}        	
    }

	/* Finish last pixel sequence */
    if (runCount > 0)	
    { 
		//D(bug("targa.datatype/WriteRLE() --- End RunCount: %d\n", runCount));
        //RLE pixel //write pixel
        outBuf[offset] = 0x80 | runCount;
        memcpy(outBuf+offset+1, pixBuf1, bpp);
        cnt = bpp + 1;
        offset += cnt;
    }
    else if (dirCount > 0)
    {        
		//D(bug("targa.datatype/WriteRLE() --- End DirCount: %d\n", dirCount));
        //Raw pixel //write pixel
        outBuf[offset] = dirCount-1;
        memcpy(outBuf+offset+1, inBuf-bpp*dirCount, bpp*dirCount);
        cnt = bpp*dirCount + 1;
        offset += cnt;
    }
        

    D(bug("targa.datatype/WriteRLE() --- Last Offset: %d\n", offset));
    

	if( FWrite(file, outBuf, offset, 1 ) != 1 )	
	{
		D(bug("targa.datatype/WriteRLE() --- writing compressed buffer failed!\n"));			
		offset = 0;
	}

    return offset;
}

LONG ConvertCmap(UBYTE *inbuf, UBYTE *outbuf, LONG len, LONG bytes)
{
	int x;
	BYTE r,g,b;
	LONG error = OK;

    /* Convert Cmap to 24bit. */	
	if(bytes == 15 || bytes == 16) 
	{	
        for(x = 0; x < len; x++)		
        {            
            /* 16 to 24 -- convert values. */            
			r = ((inbuf[(x*2)+1] << 1) & 0xf8);
			r += r >> 5;
			g = ((inbuf[x*2] & 0xe0) >> 2) | ((inbuf[(x*2)+1] & 0x03) << 6);
			g += g >> 5;
            b = ((inbuf[x*2] << 3) & 0xf8);
			b += b >> 5;

            /* Assign new values to rgb line buffer */            
            outbuf[x*3] = b; outbuf[(x*3)+1] = g; outbuf[(x*3)+2] = r;
        }
	}
	else if(bytes == 24)
	{
		for (x = 0; x < len; x++)
		{	
			/* 24 to 24 -- convert buffer. */ 
			b = inbuf[(x*3)];
			g = inbuf[(x*3) + 1];
			r = inbuf[(x*3) + 2];			
			outbuf[(x*3)] = b;
			outbuf[(x*3) + 1] = g;
			outbuf[(x*3) + 2] = r;							
		}
	}	
	else if(bytes == 32)
	{
		for (x = 0; x < len; x++)
		{	
			/* 32 to 24 -- discard alpha. */ 
			b = inbuf[(x*4)];
			g = inbuf[(x*4) + 1];
			r = inbuf[(x*4) + 2];
			//a = inbuf[(x*4) + 3];
			outbuf[(x*3)] = b;
			outbuf[(x*3) + 1] = g;
			outbuf[(x*3) + 2] = r;
			//outbuf[(x*3) + 3] = a;				
		}
	}
	else 
	   error = IoErr();

	return error;
}

LONG GetCMAP24 (struct Library *libBase, BPTR file, struct ColorRegister *cmap, LONG len, LONG bytes)
{
	LONG i;	
	UBYTE *buf, *cbuf, *ptr;
	LONG status, error = OK; 
	LONG size = len * 3;
	LONG csize = len * (bytes/8);	

	buf = AllocMem(size, MEMF_ANY);
	cbuf = AllocMem(csize, MEMF_ANY);
	if(!buf || !cbuf)	
	{
		return ERROR_NO_FREE_STORE;
	}

	status = Read(file, cbuf, csize);
	if(status != csize)	 
	{
		error = ReadError(status);
		goto out;
	}

	/* Convert cbuf to 24bit */	 
	error = ConvertCmap(cbuf, buf, len, bytes);

	
	ptr = buf;
	for (i = 0; i < len; i++) {
		cmap->blue = *ptr++;
		cmap->green = *ptr++;
		cmap->red = *ptr++;
		cmap++;
	}

    out:
	FreeMem(buf, size);
	FreeMem(cbuf, csize);
	return error;
}

LONG SetCMAP24(BPTR file, struct ColorRegister *colormap, UBYTE *CMAP, LONG len)
{
	LONG i;
	UBYTE *ptr;
	LONG size = len * 3;	

	ptr = CMAP;
	for (i = 0; i < len; i++) 
	{
		*ptr++ = colormap->blue;
		*ptr++ = colormap->green;
		*ptr++ = colormap->red;
		colormap++;
	}

	return 0;
}

static BOOL GetTarga (Class *cl, Object *o)
//static LONG GetTarga (Class *cl, Object *o, struct TagItem *tags) 
{
	//struct Library *libBase = (struct Library *)cl->cl_UserData;
	struct BitMapHeader *bmh = NULL;
	//char *filename;
	char *name;
	IPTR srctype;
	LONG error = ERROR_OBJECT_NOT_FOUND;
	BPTR file = BNULL;

    /* Pass attributes to picture.datatype */
    GetDTAttrs( o, DTA_Name, (IPTR)&name, TAG_DONE );
	//filename = (char *)GetTagData(DTA_Name, (IPTR)"Untitled", tags);

	GetDTAttrs(o,
		PDTA_BitMapHeader,	&bmh,
		DTA_Handle,			&file,
		DTA_SourceType,		&srctype,
		TAG_END);

	/* Do we have everything we need? */
	if (bmh && file && srctype == DTST_FILE) 
	{
		error = ConvertTarga(cl, o, file, bmh);
		if (error == OK) {
			SetDTAttrs(o, NULL, NULL,
				DTA_ObjName,	FilePart(name),
				TAG_END);
		}
	}
	//return error;
	return TRUE;
}
